@@ -10,6 +10,7 @@ import os
 import sys
 
 from xlsxwriter.utility import xl_rowcol_to_cell
+from functools import cmp_to_key
 
 devMinThreshold = 1.5
 devMaxThreshold = 2.0
@@ -17,14 +18,35 @@ thresholdPercent = 0.15
 
 class GenerateSpreadsheet(object):
 
-    def __init__(self, excelFile, files):
+    def __init__(self, excelFile, files, args):
 
         s = xlsxwriter.Workbook(excelFile)
 
         # keep a list of sheets so we can insert a summary..
         geninfoSheets = []
-        summarySheet = s.add_worksheet("geninfo_summary")
-        geninfoKeys = ('process', 'child', 'parse', 'append', 'exec', 'merge', 'undump')
+        summarySheet = s.add_worksheet("geninfo_summary") if 1 < len(files) else None
+
+        # order:  order of processing
+        # file: time to process one GCDA file
+        # parse:  time to generate and read gcov data
+        # exec: time to execute gcov
+        # append: to merge file info into parent
+        geninfoKeys = ('order', 'file', 'parse', 'exec', 'append')
+
+        # work: productive time: process_one_chunk + merge chunk
+        # chunk: everything from fork() to end of filesystem cleanup after child merge
+        # child: time from entering child process to immediately before serialize
+        # process: time to call process_one_chunk
+        # undump:  time to deserialize chunk data into master
+        # queue: time between child finish and start of merge in parent
+        # merge: time to merge returned chunk info
+        geninfoChunkKeys = ('work', 'chunk', 'queue', 'child', 'process', 'undump', 'merge')
+        geninfoSpecialKeys = ('total', 'parallel', 'filter', 'write')
+
+        # keys related to filtering
+        filterKeys = ('filt_chunk', 'filt_queue',  'filt_child', 'filt_proc', 'filt_undump', 'filt_merge', 'derive_end')
+        if args.verbose:
+            geninfoKeys.extend(['read', 'translate'])
 
         self.formats = {
             'twoDecimal': s.add_format({'num_format': '0.00'}),
@@ -50,7 +72,7 @@ class GenerateSpreadsheet(object):
             devCell = xl_rowcol_to_cell(devRow, beginCol, True, False)
             # relative row, relative column
             dataCell = xl_rowcol_to_cell(beginRow, beginCol, False, False)
-            # absolute value of differnce from the average
+            # absolute value of difference from the average
             diff = 'ABS(%(cell)s - %(avg)s)' % {
                 'cell' : dataCell,
                 'avg' : avgCell,
@@ -115,6 +137,8 @@ class GenerateSpreadsheet(object):
             col -= 1
             for key in keys:
                 col += 1
+                if key in ('order',):
+                    continue
                 f = xl_rowcol_to_cell(beginRow, col)
                 t = xl_rowcol_to_cell(endRow, col)
 
@@ -195,6 +219,15 @@ class GenerateSpreadsheet(object):
                     # old file format..skip it
                     pass
 
+            if tool == 'geninfo':
+                for k in ('chunkSize', 'nChunks', 'nFiles', 'interval'):
+                    try:
+                        sheet.write_number(row, 2, data[k], intFormat);
+                        sheet.write_string(row, 1, k)
+                        row += 1
+                    except:
+                        pass
+
             for k in ('total', 'overall'):
                 if k in data:
                     sheet.write_string(row, 0, 'total')
@@ -281,160 +314,233 @@ class GenerateSpreadsheet(object):
 
             elif tool == 'geninfo':
 
-                if len(geninfoSheets) == 0:
+                summaryKeys = (*geninfoSpecialKeys, *geninfoChunkKeys, *geninfoKeys)
+                if args.show_filter:
+                    summaryKeys = (*geninfoSpecialKeys, *geninfoChunkKeys, *geninfoKeys, *filterKeys)
+                if summarySheet:
                     # first one - add titles, etc
                     title = self.formats['title']
-                    summarySheet.write_string(1, 0, "average", title)
-                    summarySheet.write_string(2, 0, "stddev", title)
-                    titleRow = 0
-                    summarySheet.write_string(titleRow, 0, "case", title)
-                    col = 1
-                    for k in ('total', *geninfoKeys):
-                        summarySheet.write_string(titleRow, col, k, title)
-                        col += 1
-                        if k == 'total':
-                            summarySheet.write_string(titleRow, col, 'parallel', title)
-                        else:
+
+                    if len(geninfoSheets) == 0:
+                        summarySheet.write_string(1, 0, "average", title)
+                        summarySheet.write_string(2, 0, "stddev", title)
+                        titleRow = 0
+                        summarySheet.write_string(titleRow, 0, "case", title)
+                        col = 1
+                        for k in summaryKeys:
+                            if k in ('order',):
+                                continue
+                            summarySheet.write_string(titleRow, col, k, title)
+                            col += 1
+                            if k in geninfoSpecialKeys:
+                                continue
                             summarySheet.write_string(titleRow, col, k + ' avg', title)
-                        col += 1
-                    summarySheet.write_string(3, 0, "YELLOW: Value between [%(devMinThresh)0.2f,%(devMaxThresh)0.2f) standard deviations larger than average" % {
-                        'devMinThresh': devMinThreshold,
-                        'devMaxThresh': devMaxThreshold,
-                    }, self.formats['highlight'])
-                    summarySheet.write_string(4, 0, "RED: Value more than %(devMaxThresh)0.2f standard deviations larger than average" % {
-                        'devMaxThresh': devMaxThreshold,
-                    }, self.formats['danger'])
-                    summarySheet.write_string(5, 0, "GREEN: Value more than %(devMaxThresh)0.2f standard deviations smaller than average" % {
-                        'devMaxThresh': devMaxThreshold,
-                    }, self.formats['good'])
-                    firstSummaryRow = 7
+                            col += 1
+                        summarySheet.write_string(3, 0, "YELLOW: Value between [%(devMinThresh)0.2f,%(devMaxThresh)0.2f) standard deviations larger than average" % {
+                            'devMinThresh': devMinThreshold,
+                            'devMaxThresh': devMaxThreshold,
+                        }, self.formats['highlight'])
+                        summarySheet.write_string(4, 0, "RED: Value more than %(devMaxThresh)0.2f standard deviations larger than average" % {
+                            'devMaxThresh': devMaxThreshold,
+                        }, self.formats['danger'])
+                        summarySheet.write_string(5, 0, "GREEN: Value more than %(devMaxThresh)0.2f standard deviations smaller than average" % {
+                            'devMaxThresh': devMaxThreshold,
+                        }, self.formats['good'])
+                        firstSummaryRow = 7
+
+                    # want rows for average and variance - leave a blank row
+                    summaryRow = firstSummaryRow + len(geninfoSheets)
 
-                # want rows for average and variance - leave a blank row
-                summaryRow = firstSummaryRow + len(geninfoSheets)
                 geninfoSheets.append(sheet)
-
-                d = data['gen_info']
-                for k in ('emit', ):
+                # already inserted the ''total' entry
+                specialsStart = row - 1
+                for k in geninfoSpecialKeys[1:]:
                     try:
-                        sheet.write_number(row, 3, data[k], twoDecimal)
-                        sheet.write_string(row, 2, k)
-                        row += 1
+                        sheet.write_string(row, 0, k)
+                        sheet.write_number(row, 1, data[k], twoDecimal)
                     except:
                         pass
+                    row += 1;
 
                 sawData = {}
                 sawData['total'] = 0
-                summarySheet.write_string(summaryRow, 0, name)
-                # href to the corresponding page..
-                summarySheet.write_url(summaryRow, 0, "internal:'%s'!A1" % (
-                    sheet.get_name()))
-                summaryCol = 1;
-
-                sheetRef = "='" + sheet.get_name() + "'!"
-
-                # insert total time and observed parallelism for this
-                # geninfo call
-                sum = xl_rowcol_to_cell(totalRow, 1)
-                summarySheet.write_formula(summaryRow, summaryCol,
-                                           sheetRef + sum)
-                summaryCol += 1
-                parallel = xl_rowcol_to_cell(totalRow, 2)
-                summarySheet.write_formula(summaryRow, summaryCol,
-                                           sheetRef + parallel)
-                summaryCol += 1
-                col = 3;
-                sheet.write_string(row, col, 'cumulative')
-                col += 1
-                # now label this sheet's columns
-                #  and also insert reference to total time and average time
-                #  for each step into the summary sheet.
-                for k in geninfoKeys:
-                    sheet.write_string(row, col, k)
-                    sum = xl_rowcol_to_cell(row+1, col)
-                    summarySheet.write_formula(summaryRow, summaryCol,
-                                               sheetRef + sum)
-                    summaryCol +=1
-                    avg = xl_rowcol_to_cell(row+2, col)
-                    summarySheet.write_formula(summaryRow, summaryCol,
-                                               sheetRef + avg)
-                    summaryCol +=1
-                    col += 1
+                sheet.write_string(row, 0, 'find')
                 row += 1
-                sumRow = row
-                sheet.write_string(row, 2, "total")
-                row += 1
-                avgRow = row
-                sheet.write_string(row, 2, "average")
-                row += 1
-                devRow = row
-                sheet.write_string(row, 2, "stddev")
-                row += 1
-
-                firstRow = row + 2
-                for dirname in sorted(d.keys()):
-                    sheet.write_string(row, 0, 'gen_info')
+                for dirname in sorted(data['find'].keys()):
                     sheet.write_string(row, 1, dirname)
-                    sheet.write_number(row, 2, d[dirname], twoDecimal)
+                    sheet.write_number(row, 2, data['find'][dirname], twoDecimal)
                     row += 1
 
-                    start = row
-                    sheet.write_string(row, 2, 'find')
-                    sheet.write_number(row, 3, data['find'][dirname], twoDecimal)
+                row += 1
+                def dataSection(typename, elements, keylist, dataRow, statsRow):
+
+                    row = dataRow
+                    sheet.write_string(row, 0, typename)
+                    col = 2
+                    for key in keylist:
+                        sheet.write_string(row, col, key)
+                        col += 1
                     row += 1
+                    dataStart = row
 
-                    for type in ('data', 'graph'):
-                        if type not in data:
-                            continue
+                    sawData = {}
+                    for id in elements:
+                        col = 1
+                        sheet.write_string(row, col, id)
+                        col += 1
 
-                        d2 = data[type][dirname]
-                        for f in sorted(d2.keys()):
-                            d3 = d2[f]
-                            fname = f[len(dirname):]
-                            sheet.write_string(row, 2, fname)
+                        for key in keylist:
                             try:
-                                # this is the total time from fork in the parent
-                                #  to end of child merge
-                                sheet.write_number(row, 3, float(d3), twoDecimal)
-                                sawData['total'] += 1
-                            except:
-                                print("%s: failed to write %s for geninfo[%s][%s]" % (name, str(d3), type, f))
-                            col = 4
-                            # process: time from immediately before fork in parent
-                            #          to immediately after 'process_one_file' in
-                            #          child (can't record 'dumper' call time
-                            #          because that also dumps the profile
-                            # child:   time from child coming to life after fork
-                            #          to immediately afer 'process_one_file'
-                            # exec: time take to by 'gcov' call
-                            # merge: time to merge child process (undump, read
-                            #       trace data, append to summary, etc.)
-                            # undump: dumper 'eval' call + stdout/stderr recovery
-                            # parse: time to read child tracefile.info
-                            # append: time to merge that into parent master report
-
-                            for key in geninfoKeys:
+                                v = data[key][id]
+                                if key in ('order',):
+                                    sheet.write_number(row, col, v, intFormat)
+                                else:
+                                    sheet.write_number(row, col, v, twoDecimal)
                                 try:
-                                    val = float(data[key][dirname][f])
-                                    sheet.write_number(row, col, val, twoDecimal)
-                                    if key in sawData:
-                                        sawData[key] += 1
-                                    else:
-                                        sawData[key] = 1
+                                    sawData[key] += 1
                                 except:
-                                    pass # no such key
-                                col += 1
-                            row += 1
+                                    sawData[key] = 1
+
+                            except:
+                                pass
+                            col += 1
+                        row += 1
+
+                    dataEnd = row - 1
+
+                    row = statsRow
+                    # insert link to first associated data entry
+                    sheet.write_url(row, 0, "internal:'%s'!%s" %(
+                        sheet.get_name(),
+                        xl_rowcol_to_cell(dataStart, 0)),
+                                    string=typename)
+                    col = 2
+                    for key in keylist:
+                        if key not in ('order', ):
+                            sheet.write_string(row, col, key)
+                        col += 1
+                    row += 1
+                    sheet.write_string(row, 1, 'total')
+                    sheet.write_string(row+1, 1, 'avg')
+                    sheet.write_string(row+2, 1, 'stddev')
+                    insertStats(keylist, sawData, statsRow + 1, statsRow + 2,
+                                statsRow+3, dataStart, dataEnd, 2)
+                    return dataEnd + 1
+
+                chunkStatsRow = row
+                fileStatsRow = chunkStatsRow + 4
+                filterStatsRow = fileStatsRow + 4;
+
+                if args.show_filter:
+                    chunkDataRow = filterStatsRow + 4
+                else:
+                    chunkDataRow = fileStatsRow + 4
+                    fileStatsRow = row
+                parallelSumRow = row+1
+                parallelSumCol = 3
+
+                # first the chunk data...
+                # process: time from immediately before fork in parent
+                #          to immediately after 'process_one_file' in
+                #          child (can't record 'dumper' call time
+                #          because that also dumps the profile
+                # child:   time from child coming to life after fork
+                #          to immediately after 'process_one_file'
+                # exec: time take to by 'gcov' call
+                # merge: time to merge child process (undump, read
+                #       trace data, append to summary, etc.)
+                # undump: dumper 'eval' call + stdout/stderr recovery
+                # parse: time to read child tracefile.info
+                # append: time to merge that into parent master report
+                try:
+                    chunks = sorted(data['child'].keys(), key=int, reverse=True)
+                    row = dataSection('chunks', chunks, geninfoChunkKeys,
+                                      chunkDataRow, chunkStatsRow)
+                    row += 1
+                    fileDataRow = row + 1
+                    parallelSumCol = 2
+                except:
+                    # no chunk data - so just insert file data
+                    fileStatsRow = chunkStatsRow
+                    fileDataRow = fileStatsRow + 4
+                    pass
+
+
+                def cmpFile(a, b):
+                    idA = int(data['order'][a])
+                    idB = int(data['order'][b])
+                    if idA < idB:
+                        return 1
+                    else:
+                        return 0 if idA == idB else -1
+
+                row = dataSection('files', sorted(data['file'].keys(), key=cmp_to_key(cmpFile)),
+                                  geninfoKeys, fileDataRow, fileStatsRow)
+
+                # now the filter data - if any
+                if args.show_filter:
+                    filterDataRow = row + 1;
+                    try:
+                        chunks = sorted(data['filt_child'].keys(), key=int, reverse=True)
+                        row = dataSection('filter', chunks, filterKeys,
+                                          filterDataRow, filterStatsRow)
+                        row += 1
+
+                    except:
+                        pass
+
 
                 effectiveParallelism = "+%(sum)s/%(total)s" % {
-                    'sum': xl_rowcol_to_cell(sumRow, 4),
+                    'sum': xl_rowcol_to_cell(parallelSumRow, parallelSumCol),
                     'total': total,
                 }
-                sheet.write_formula(totalRow, 2, effectiveParallelism, twoDecimal)
-                keys = ['total']
-                keys.extend(geninfoKeys)
-                insertStats(keys, sawData, sumRow, avgRow, devRow,
-                            firstRow, row-1, 3)
+                sheet.write_formula(specialsStart + geninfoSpecialKeys.index('parallel'),
+                                    1, effectiveParallelism, twoDecimal)
 
+                if summarySheet:
+                    summarySheet.write_string(summaryRow, 0, name)
+                    # href to the corresponding page..
+                    summarySheet.write_url(summaryRow, 0, "internal:'%s'!A1" % (
+                        sheet.get_name()))
+                    summaryCol = 1;
+
+                    sheetRef = "='" + sheet.get_name() + "'!"
+
+                    # insert total time and observed parallelism for this
+                    # geninfo call
+                    specialsRow = specialsStart
+                    for k in geninfoSpecialKeys:
+                        cell  = xl_rowcol_to_cell(specialsRow, 1)
+                        summarySheet.write_formula(summaryRow, summaryCol,
+                                                   sheetRef + cell, twoDecimal)
+                        summaryCol += 1
+                        specialsRow += 1
+
+                    # now label this sheet's columns
+                    #  and also insert reference to total time and average time
+                    #  for each step into the summary sheet.
+                    statsTotalRow = chunkStatsRow + 1
+                    statsAvgRow = chunkStatsRow + 2
+                    sections = [(geninfoChunkKeys, chunkStatsRow + 1, chunkStatsRow + 2),
+                              (geninfoKeys, fileStatsRow + 1, fileStatsRow + 2),]
+                    if args.show_filter:
+                        sections.append((filterKeys, filterStatsRow +1,
+                                         filterStatsRow+2))
+                    for d in (sections):
+                        totRow = d[1]
+                        avgRow = d[2]
+                        col = 2
+                        for k in d[0]:
+                            if k not in ('order',):
+                                sum = xl_rowcol_to_cell(totRow, col)
+                                summarySheet.write_formula(summaryRow, summaryCol,
+                                                           sheetRef + sum, twoDecimal)
+                                summaryCol +=1
+                                avg = xl_rowcol_to_cell(avgRow, col)
+                                summarySheet.write_formula(summaryRow, summaryCol,
+                                                           sheetRef + avg, twoDecimal)
+                                summaryCol +=1
+                            col += 1
                 continue
 
             elif tool == 'genhtml':
@@ -508,7 +614,7 @@ class GenerateSpreadsheet(object):
                 try:
                     dirData = data['dir']
                 except:
-                    # hack - 'flat' report doens't have directory data
+                    # hack - 'flat' report doesn't have directory data
                     for f in sorted(fileData.keys()):
                       visitScope(f, None)
                     dirData = {}
@@ -560,35 +666,40 @@ class GenerateSpreadsheet(object):
                 else:
                     print("not sure what to do with %s" % (k))
 
-        if len(geninfoSheets) < 2:
-            # can't delete the sheet after creation - but we can hide it
-            summarySheet.hide()
-        else:
-            # insert the average and variance data...
-            col = 1
-            lastSummaryRow = firstSummaryRow + len(geninfoSheets) - 1
-            avgRow = 1
-            devRow = 2
-            firstCol = col
-            for k in ('total', *geninfoKeys):
-                for j in ('sum', 'avg'):
-                    f = xl_rowcol_to_cell(firstSummaryRow, col)
-                    t = xl_rowcol_to_cell(lastSummaryRow, col)
-                    avg = "+AVERAGE(%(from)s:%(to)s)" % {
-                        'from': f,
-                        'to': t,
-                    }
-                    summarySheet.write_formula(avgRow, col, avg, twoDecimal)
-                    avgCell = xl_rowcol_to_cell(avgRow, col)
-                    dev = "+STDEV(%(from)s:%(to)s)" % {
-                        'from': f,
-                        'to': t,
-                    }
-                    summarySheet.write_formula(devRow, col, dev, twoDecimal)
-                    col += 1
+        if summarySheet:
+            if len(geninfoSheets) < 2:
+                summarySheet.hide()
 
-            insertConditional(summarySheet, avgRow, devRow,
-                              firstSummaryRow, firstCol, lastSummaryRow, col -1)
+            # insert the average and variance data...
+            #  (there will not be any such data if we didn't run geninfo)
+            try:
+                col = 1
+                lastSummaryRow = firstSummaryRow + len(geninfoSheets) - 1
+                avgRow = 1
+                devRow = 2
+                firstCol = col
+                for k in (*geninfoChunkKeys, *geninfoKeys):
+                    if k in ('order',):
+                        continue
+                    for j in ('sum', 'avg'):
+                        f = xl_rowcol_to_cell(firstSummaryRow, col)
+                        t = xl_rowcol_to_cell(lastSummaryRow, col)
+                        avg = "+AVERAGE(%(from)s:%(to)s)" % {
+                            'from': f,
+                            'to': t,
+                        }
+                        summarySheet.write_formula(avgRow, col, avg, twoDecimal)
+                        avgCell = xl_rowcol_to_cell(avgRow, col)
+                        dev = "+STDEV(%(from)s:%(to)s)" % {
+                            'from': f,
+                            'to': t,
+                        }
+                        summarySheet.write_formula(devRow, col, dev, twoDecimal)
+                        col += 1
+                insertConditional(summarySheet, avgRow, devRow,
+                                  firstSummaryRow, firstCol, lastSummaryRow, col -1)
+            except:
+                pass
         s.close()
 
 if __name__ == "__main__":
@@ -610,6 +721,10 @@ Example usage:
                         help="difference from average larger than this * stddev colored yellow.  Default: %0.2f" %(devMinThreshold))
     parser.add_argument("--high", dest='devMinThreshold', type=float,
                         help="difference from average larger than this * stddev colored red.  Default: %0.2f" %(devMaxThreshold))
+    parser.add_argument('-v', '--verbose', dest='verbose', default=0,
+                        action='count', help='verbosity of report: more data');
+    parser.add_argument('--show-filter', dest='show_filter', default=False,
+                        action='store_true', help='include filter keys in table');
 
     parser.add_argument('files', nargs=argparse.REMAINDER)
 
@@ -619,5 +734,4 @@ Example usage:
         print(str(err))
         sys.exit(2)
 
-
-    GenerateSpreadsheet(args.out, args.files)
+    GenerateSpreadsheet(args.out, args.files, args)
